--=====================================================================================================================
--== ScanDoubler pro Amstrad/Schneider CPC								      DL2011 ==
--=====================================================================================================================
--== verze 1.19												 	     ==
--=====================================================================================================================
--== Zdrojovy kod je mi uplne volny :-))									     ==
--== Veskere sireni a pozmenovani je povoleno									     ==
--=====================================================================================================================

library IEEE;
use IEEE.std_logic_1164.ALL;
use IEEE.std_logic_arith.ALL;
use IEEE.std_logic_unsigned.ALL;
entity SCANDOUB is port (
-----------------------------------------------------------------------------------------------------------------------
-- vnejsi piny pro pripojeni k Sharp CPC ------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
KCLK		: in std_logic;					--hodiny 16 MHz
KRGB		: in std_logic_vector(5 downto 0);		--barevne slozky
KHSYNC		: in std_logic;					--horizontalni synchronizace
KVSYNC		: in std_logic;					--vertikalni synchronizace
KSCAN		: in std_logic;					--invertuje barvy na vystupu
KSTOP		: in std_logic;					--zapauzuje vga obraz
-----------------------------------------------------------------------------------------------------------------------
-- vnejsi piny pro pripojeni RAM --------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
RAM_AD		: out std_logic_vector(17 downto 0);		--adresni sbernice
RAM_DT		: inout std_logic_vector(15 downto 0);		--datova sbernice
RAM_WR		: out std_logic;				--povoleni zapisu
-----------------------------------------------------------------------------------------------------------------------
-- vnejsi pin pro vstup hodin 50MHz -----------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
CLK50		: in std_logic;					--hodiny 50MHz
-----------------------------------------------------------------------------------------------------------------------
-- vnejsi piny pro pripojeni k VGA ------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
VH_RED		: out std_logic;				--primarni cervena slozka
VL_RED		: out std_logic;				--sekundarni cervena slozka
VH_GREEN	: out std_logic;				--primarni zelena slozka
VL_GREEN	: out std_logic;				--sekundarni zelena slozka
VH_BLUE		: out std_logic;				--primarni modra slozka
VL_BLUE		: out std_logic;				--sekundarni zelena slozka
V_INTENS	: out std_logic;				--pridavna intenzita pro vsechny barevne slozky
V_HSYNC		: out std_logic;				--horizontalni synchronizace
V_VSYNC		: out std_logic);				--vertikalni synchronizace
-----------------------------------------------------------------------------------------------------------------------
end SCANDOUB;
architecture rtl of SCANDOUB is
-----------------------------------------------------------------------------------------------------------------------
-- vnitrni signaly pro vstupni cast -----------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
signal rKHSYNC	: std_logic;					--pamet stavu signalu KHSYNC
signal rKVSYNC	: std_logic;					--pamet stavu signalu KVSYNC
signal kCTCHOR	: std_logic_vector(9 downto 0);			--horizontalni citac vstupniho obrazu
signal kCTCVER	: std_logic_vector(8 downto 0);			--vertikalni citac vstupniho obrazu
signal kPAGE	: std_logic;					--vyber stranky do ktere se zapisuje
signal k1PIX	: std_logic_vector(5 downto 0);			--pamet pro pixel na vstupu
signal k2PIX	: std_logic_vector(11 downto 0);		--pamet pro 2 pixely na vstupu (slovo pro zapis do RAM)
-----------------------------------------------------------------------------------------------------------------------
-- vnitrni signaly pro VGA cast ---------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
signal vgaCTCH	: std_logic_vector(10 downto 0);		--VGA horizontalni citac
signal vgaCTCV	: std_logic_vector(9 downto 0);			--VGA vertikalni citac
signal vgaENAB	: std_logic;					--povoleni zobrazeni
signal vgaENAv	: std_logic;					--povoleni zobrazeni od vertikalniho citace
signal vgaPAGE	: std_logic;					--vyber stranky ze ktere se zobrazuje
signal vga2PIX	: std_logic_vector(11 downto 0);		--pamet pro 2 pixely na vystupu (slovo prectene z RAM)
signal vgaPIX	: std_logic_vector(5 downto 0);			--barevny vystup jeden pixel
-----------------------------------------------------------------------------------------------------------------------
-- vnitrni signaly pro zapis do pameti RAM ----------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
signal kWRREQ	: std_logic;					--pozadavek zapisu (vstupni cast)
signal rWRREQ	: std_logic_vector(1 downto 0);			--register pozadavku zapisu (pro synchronizaci)
signal pWRREQ	: std_logic;					--pozadavek zapisu (VGA cast)
signal iRAMWR	: std_logic;					--interni signal WR pro pamet RAM
-----------------------------------------------------------------------------------------------------------------------
begin
-----------------------------------------------------------------------------------------------------------------------
--citace a synchronizace dat na vstupni casti -------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
process (KCLK) begin
    if KCLK'event and KCLK='1' then
	rKHSYNC <= KHSYNC;					--pamet stavu KHSYNC
	rKVSYNC <= KVSYNC;					--pamet stavu KVSYNC
	kWRREQ <= '0';						--pozadavek o zapis je defaultne neaktivni
	if rKHSYNC='0' and KHSYNC='1' then			--detekovan konec aktivity KHSYNC
	    kCTCHOR <= "1100100010";				--inicializace horizontalniho citace (-/+ obraz vlevo/vpravo)
	    kCTCVER <= kCTCVER + 1;				--vertikalni citac +1
	elsif kCTCHOR(0)='0' then
	    k1PIX <= KRGB;					--registruj prvni pixel
	    kCTCHOR(0) <= '1';					--citac na dalsi pixel
	else
	    k2PIX <= KRGB & k1PIX;				--registruj vsechny ctyri pixely
	    kCTCHOR <= kCTCHOR + '1';				--citac na dalsi pixel
	    kWRREQ <= not kCTCVER(8);				--nastav pozadavek o zapis
	end if;
	if rKVSYNC='0' and KVSYNC='1' then			--detekovana nabezna hrana KVSYNC
	    kCTCVER <= "111010100";				--inicializace vertikalniho citace (-/+ obraz nahoru/dolu)
	    if KSTOP='1' then					--pouze pokud neni stop stav
	 	kPAGE <= not kPAGE;				--preklopeni do druhe poloviny pameti
	    end if;
	end if;
    end if;
end process;
-----------------------------------------------------------------------------------------------------------------------
process (CLK50) begin
    if CLK50'event and CLK50='0' then
-----------------------------------------------------------------------------------------------------------------------
-- detekce pozadavku o zapis do pameti RAM od vstupni casti -----------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
	if rWRREQ=2 then					--detekovan pozadavek o zapis ze vstupni casti
	    pWRREQ <= '1';					--nastav pozadavek o zapis
	end if;
	rWRREQ <= kWRREQ & rWRREQ(1);				--registruj pozadavek o zapis ze vstupni casti
-----------------------------------------------------------------------------------------------------------------------
-- generovani signalu WR pro pamet RAM --------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
	iRAMWR <= '1';						--zapis je defaultne neaktivni
	if pWRREQ='1' and vgaCTCH(0)='1' then			--pokud je pozadavek zapisu, ktery je mozno vykonat
	    iRAMWR <= '0';					--nastav vnitrni signal /WR
	    pWRREQ <= '0';					--zrus pozadavek zapisu
	end if;
-----------------------------------------------------------------------------------------------------------------------
-- VGA - citani, nulovani horizontalniho a vertikalniho citace --------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
	if vgaCTCH=1039 then					--pokud jsme docitali sloupce na radku
	    vgaCTCH <= (others=>'0');				--nulovani horizontalniho citace
	    if vgaCTCV=665 then					--pokud jsme docitali radky na strance
		vgaCTCV <= (others=>'0');			--nulovani vertikalniho citace
	    else
		vgaCTCV <= vgaCTCV + '1';			--vertikalni citac +1
		if vgaCTCV=621 then
		    vgaPAGE <= not kPAGE;			--preklopeni do pameti do ktere se nezapisuje
		end if;
	    end if;
	else
	    vgaCTCH <= vgaCTCH + '1';				--horizontalni citac +1
	end if;
-----------------------------------------------------------------------------------------------------------------------
-- VGA - generovani vertikalni synchronizace --------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
	if vgaCTCV=622 then
	    vgaENAv <= '1';					--povoleni zobrazeni
	elsif vgaCTCV=556 then
	    vgaENAv <= '0';					--zakazani zobrazeni
	elsif vgaCTCV=593 then
	    V_VSYNC <= '1';					--zacatek vertikalni synchronizace
	elsif vgaCTCV=599 then
	    V_VSYNC <= '0';					--konec vertikalni synchronizace
	end if;
-----------------------------------------------------------------------------------------------------------------------
-- VGA - generovani horizontalni synchronizace ------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
	if vgaCTCH=1 and vgaENAv='1' then
	    vgaENAB <= '1';					--povoleni zobrazovani
	elsif vgaCTCH=801 then
	    vgaENAB <= '0';					--zakazani zobrazovani
	elsif vgaCTCH=857 then
	    V_HSYNC <= '1';					--zacatek horizontalni synchronizace
	elsif vgaCTCH=977 then
	    V_HSYNC <= '0';					--konec horizontalni synchronizace
	end if;
-----------------------------------------------------------------------------------------------------------------------
-- VGA - cteni dat z pameti a jejich zapis do registru, rotace registru -----------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
	if vgaCTCH(0)='1' then
	    vga2PIX <= RAM_DT(11 downto 0);			--zapis prectenych dat z pameti do registru
	else
	    vga2PIX(5 downto 0) <= vga2PIX(11 downto 6);	--posunuti registru na dalsi pixel
	end if;
-----------------------------------------------------------------------------------------------------------------------
    end if;
end process;
-----------------------------------------------------------------------------------------------------------------------
-- VGA - generovani barvonosnych signalu ------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
vgaPIX <=	"000000" when vgaENAB='0' else			--zobrazovani zakazano
		vga2PIX(5 downto 0);				--barvy pixelu
VH_RED <=	vgaPIX(0);					--primarni cervena slozka
VL_RED <=	'0' when KSCAN='0' and vgaCTCV(0)='1' else	--sekundarni cervena slozka
		vgaPIX(3);
VH_GREEN <=	vgaPIX(1);					--primarni zelena slozka
VL_GREEN  <=	'0' when KSCAN='0' and vgaCTCV(0)='1' else	--sekundarni zelena slozka
		vgaPIX(4);
VH_BLUE <=	vgaPIX(2);					--primarni modra slozka
VL_BLUE  <=	'0' when KSCAN='0' and vgaCTCV(0)='1' else	--sekundarni modra slozka
		vgaPIX(5);
V_INTENS <=	'0' when vgaPIX=0 else				--nepouzije se pri cerne barve a pri zatemneni
		'0' when KSCAN='1' else				--nepouzije se pokud jsou scanlines zakazany
		'0' when vgaCTCV(0)='1' else			--nepouzije se u scanlines na kazdem lichem radku
		'1';
-----------------------------------------------------------------------------------------------------------------------
-- napojeni pameti RAM ------------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------
RAM_AD <= kPAGE & kCTCVER(7 downto 0) & kCTCHOR(9 downto 1) when iRAMWR='0' else	-- cpc zapis
	  vgaPAGE & "00000000" & vgaCTCH(9 downto 1) when vgaCTCV(9 downto 8)=3 else	-- vga cteni falesne horni
	  vgaPAGE & "11111111" & vgaCTCH(9 downto 1) when vgaCTCV(9 downto 8)=2 else	-- vga cteni falesne dolni
	  vgaPAGE & vgaCTCV(8 downto 1) & vgaCTCH(9 downto 1);				-- vga cteni
RAM_DT <= "0000" & k2PIX when iRAMWR='0' else "ZZZZZZZZZZZZZZZZ";
RAM_WR <= iRAMWR;
-----------------------------------------------------------------------------------------------------------------------
end rtl;